###
# This file is a part of the NVDA App Module for Cakewalk Sonar.
# URL: http://launchpad.net/nvda-sonar/
# Copyright 2009-2013 James Teh, Victor Tsaran
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2.0, as published by
# the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# This license can be found at:
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
###

"""NVDA app module for Cakewalk Sonar
@author: James Teh <jamie@jantrid.net>
@author: Victor Tsaran <vtsaran@yahoo.com>
@copyright: 2009-2013 James Teh, Victor Tsaran
@license: GNU General Public License version 2.0
"""

import sys
import time
import re
import ctypes.wintypes
from collections import namedtuple
import wx
import comtypes.client
from comtypes import COMError
import winUser
from NVDAObjects import NVDAObject
from NVDAObjects import IAccessible
import speech
from keyboardHandler import KeyboardInputGesture
import keyboardHandler
import scriptHandler
import api
import queueHandler
import controlTypes
import appModuleHandler
from logHandler import log
import screenBitmap
import ui
import textInfos
import config
import IAccessibleHandler

WM_LBUTTONDOWN = 0x0201
WM_LBUTTONUP = 0x0202
WM_RBUTTONDOWN = 0x0204
WM_RBUTTONUP = 0x0205

class SonarSurface(object):

	### Constants

	# Strip types (SONAR_MIXER_STRIP)
	MIX_STRIP_TRACK= 0
	MIX_STRIP_AUX=1
	MIX_STRIP_MAIN=2
	MIX_STRIP_MIDI_SEND=3
	MIX_STRIP_BUS=4
	MIX_STRIP_MASTER =5
	MIX_STRIP_RACK = 6

	# Touch types (SONAR_MIXER_TOUCH)
	MIX_TOUCH_MANUAL=0
	MIX_TOUCH_NORMAL	= 1
	MIX_TOUCH_TIMEOUT	=2

	# Built-in filters (SONAR_MIXER_FILTER)
	MIX_FILTER_EQ= 0
	MIX_FILTER_COMP=1
	MIX_FILTER_GATE=2

	# Mixer Parameters (SONAR_MIXER_PARAM)
	MIX_PARAM_VOL = 0
	MIX_PARAM_PAN = 1
	MIX_PARAM_MUTE = 2
	MIX_PARAM_SOLO = 3
	MIX_PARAM_ARCHIVE = 4
	MIX_PARAM_RECORD_ARM = 5
	MIX_PARAM_VOL_TRIM = 6
	MIX_PARAM_PHASE = 7
	MIX_PARAM_INTERLEAVE = 8
	MIX_PARAM_INPUT_MAX = 9
	MIX_PARAM_INPUT = 10
	MIX_PARAM_OUTPUT_MAX = 11
	MIX_PARAM_OUTPUT = 12
	MIX_PARAM_IS_MIDI = 13
	MIX_PARAM_BANK = 14
	MIX_PARAM_PATCH = 15
	MIX_PARAM_KEY_OFFSET = 16
	MIX_PARAM_VEL_OFFSET = 17
	MIX_PARAM_SELECTED = 18
	MIX_PARAM_SEND_ENABLE = 19
	MIX_PARAM_SEND_MUTE = 20
	MIX_PARAM_SEND_VOL = 21
	MIX_PARAM_SEND_PAN = 22
	MIX_PARAM_SEND_PREPOST = 23
	MIX_PARAM_AUDIO_METER = 24
	MIX_PARAM_MIDI_METER = 25
	# These are obsolete
	#MIX_PARAM_PLAYBACK_VU = 24
	#MIX_PARAM_RECORD_VU = 25
	MIX_PARAM_FX = 26
	MIX_PARAM_FX_COUNT = 27
	MIX_PARAM_FX_PARAM_COUNT = 28
	MIX_PARAM_FX_PARAM = 29
	MIX_PARAM_TRACK_EXISTS = 30
	MIX_PARAM_BUS_TYPE = 31
	MIX_PARAM_FILTER = 32
	MIX_PARAM_FILTER_COUNT = 33
	MIX_PARAM_FILTER_PARAM_COUNT = 34
	MIX_PARAM_FILTER_PARAM = 35
	MIX_PARAM_INPUT_ECHO = 36
	MIX_PARAM_SENDCOUNT = 49
	MIX_PARAM_SENDNAME = 50
	MIX_PARAM_LAYER_MUTE = 53
	MIX_PARAM_LAYER_SOLO = 54
	MIX_PARAM_LAYER_COUNT = 55

	# Transport state types (SONAR_TRANSPORT_STATE)
	TRANSPORT_STATE_AUDIO = 0
	TRANSPORT_STATE_PLAY = 1
	TRANSPORT_STATE_SCRUB = 2
	TRANSPORT_STATE_REC = 3
	TRANSPORT_STATE_REC_AUTOMATION = 4
	TRANSPORT_STATE_LOOP = 5

	# Transport Time (Transport_Time)
	TIME_CURSOR = 0

	# Time formats (MFX_TIME_FORMAT)
	TF_NULL = 0
	TF_SECONDS = 1
	TF_SAMPLES = 2
	TF_TICKS = 3
	TF_UTICKS = 4
	TF_MBT = 5
	TF_FRAMES = 6
	TF_FRAMES_REL = 7
	TF_SMPTE = 8
	TF_SMPTE_REL = 9

WNDENUMPROC = ctypes.WINFUNCTYPE(ctypes.wintypes.BOOL, ctypes.wintypes.HWND, ctypes.wintypes.LPARAM)
def findDescendantWindow(parent, visible=None, controlID=None, className=None):
	# We need something mutable to store the result from the callback.
	result = []
	@WNDENUMPROC
	def callback(window, data):
		if (
			(visible is None or winUser.isWindowVisible(window) == visible)
			and (not controlID or winUser.getControlID(window) == controlID)
			and (not className or winUser.getClassName(window) == className)
		):
			result.append(window)
			return False
		return True
	ctypes.windll.user32.EnumChildWindows(parent, callback, 0)
	try:
		return result[0]
	except IndexError:
		raise LookupError("No matching descendant window found")

def objFromWindow(window):
	return IAccessible.getNVDAObjectFromEvent(window, winUser.OBJID_CLIENT, 0)

def waitForNoModifiers():
	for t in xrange(100):
		if not keyboardHandler.currentModifiers:
			return True
		# Wait for modifiers to be released.
		time.sleep(0.01)
	# Modifiers are being held, so don't do this.
	# Otherwise, we might trigger unintended actions.
	return False

Toolbar = namedtuple("Toolbar", ("controlID", "name"))
# Translators: The name of the Playback State toolbar.
PLAYBACK_STATE_TOOLBAR = Toolbar(59434, _("Playback State"))
# Translators: The name of the Automation toolbar.
AUTOMATION_TOOLBAR = Toolbar(59435, _("Automation"))
# Translators: The name of the Track View toolbar.
TRACK_VIEW_TOOLBAR = Toolbar(59392, _("Track View"))

def toggleToolbarControl(toolbar, childID, onMessage, offMessage, gesture=None):
	try:
		window = findDescendantWindow(api.getForegroundObject().windowHandle, controlID=toolbar.controlID, visible=True)
		control = IAccessible.getNVDAObjectFromEvent(window, winUser.OBJID_CLIENT, childID)
	except:
		# Translators: An error reported when a toolbar control can't be found.
		# %s is replaced with the name of the toolbar.
		ui.message(_("Couldn't find toolbar control. Ensure the %s toolbar is enabled.") % toolbar.name)
		if gesture:
			gesture.send()
		return
	oldStates = control.states
	if gesture:
		gesture.send()
	else:
		control.doAction()
	for t in xrange(10):
		control.invalidateCache()
		states = control.states
		if states != oldStates:
			break
		time.sleep(0.01)
	ui.message(onMessage if controlTypes.STATE_CHECKED in states else offMessage)

class Cwaccess(IAccessible.IAccessible):

	RE_CWACCESS = re.compile(r"(.*?)(?: ?@ ?(\d+)(\*)?(\^)?)?$")

	def _update(self):
		data = self.name
		self.stripField = self.stripID = self.stripIndex = None
		self.isSelected = self.isInQuickGroup = False

		if not data:
			return
		m = self.RE_CWACCESS.match(data)
		if not m:
			return

		self.stripField, self.stripID, isSelected, isInQuickGroup = m.groups()
		if self.stripID:
			self.stripIndex = int(self.stripID)
		if isSelected:
			self.isSelected = True
		if isInQuickGroup:
			self.isInQuickGroup = True

		if self.stripField == "Right click to patch an effect":
			# Translators: Reported for the FX bin (where effects are inserted)
			# on a strip.
			self.stripField = _("FX bin")

	def initOverlayClass(self):
		self._update()

	def event_nameChange(self):
		oldStripID = self.stripID
		oldStripField = self.stripField
		self._update()
		focus = api.getFocusObject()
		try:
			focus.event_cwaccessUpdate(stripIDChanged=(oldStripID != self.stripID), stripFieldChanged=(oldStripField != self.stripField))
		except AttributeError:
			pass

class CwaccessView(IAccessible.IAccessible):
	surfaceStripType = None
	isStripIDRelevant = True
	presentationType = NVDAObject.presType_content
	forceUpdateKeys = ("leftArrow", "rightArrow")

	@classmethod
	def getClass(cls, obj):
		if not obj.hasFocus:
			return None
		wclass = obj.windowClassName
		wcid = obj.windowControlID
		name = obj.name
		if wclass == "CTVBaseView":
			if wcid == 59680:
				return TrackPane
			elif wcid == 59696:
				return BusPane
		elif wclass.startswith("AfxFrameOrView"):
			if wcid == 59648:
				child = obj.firstChild
				if not child:
					return SynthRack
				elif child.windowClassName.startswith("AfxWnd"):
					return TrackInspectorPane
				else:
					return FXPresetToolbar
			elif wcid == 59649:
				return MatrixViewToolbar
			elif wcid == 59664:
				return MatrixViewRowControls
			elif wcid == 59665:
				return MatrixViewRow
		elif wclass == "CEventView":
			return EventView
		elif wclass == "#32770" and name and name.startswith("AudioSnap - "):
			return CwaccessView # AudioSnap Palette

	def initOverlayClass(self):
		self._cwaccess = self.appModule._cwaccess
		self._surface = self.appModule._surface
		self._shouldIgnoreCwaccessUpdate = False
		self._moveToFieldText = None

	def event_gainFocus(self):
		if scriptHandler.isScriptWaiting():
			return super(CwaccessView, self).event_gainFocus()
		# The cwaccess window doesn't update when focus changes, so force an update.
		self._shouldIgnoreCwaccessUpdate = True
		self._forceCwaccessUpdate()
		wx.CallLater(30, self._delayedGainFocus)

	def _delayedGainFocus(self):
		self._shouldIgnoreCwaccessUpdate = False
		super(CwaccessView, self).event_gainFocus()

	def _forceCwaccessUpdate(self):
		"""Force the cwaccess window to update.
		"""
		if not self.forceUpdateKeys:
			return
		if not waitForNoModifiers():
			return
		for key in self.forceUpdateKeys:
			KeyboardInputGesture.fromName(key).send()

	def _get_stripName(self):
		if not self._surface or self.surfaceStripType is None:
			return None
		try:
			return self._surface.GetMixStripName(
				self.surfaceStripType, self._cwaccess.stripIndex - 1)
		except COMError:
			pass

	def _get__stripIDAndStatesText(self):
		out = [self._cwaccess.stripID]
		if not out[0]:
			return
		if config.conf["documentFormatting"]["reportTableHeaders"]:
			name = self.stripName
			if name:
				out.append(name)
		if self._cwaccess.isSelected:
			out.append(" selected")
		if self._cwaccess.isInQuickGroup:
			out.append(" in quick group")
		return " ".join(out)

	def event_cwaccessUpdate(self, stripIDChanged, stripFieldChanged):
		if self._shouldIgnoreCwaccessUpdate:
			return
		if self._moveToFieldText:
			# We're trying to move to a field.
			self._moveToFieldNext(stripIDChanged)
			return
		if stripIDChanged and self.isStripIDRelevant:
			speech.speakText(self._stripIDAndStatesText)
		speech.speakText(self._cwaccess.stripField)

	def _get_value(self):
		if controlTypes.STATE_FOCUSED not in self.states or not (self._cwaccess.stripID or self._cwaccess.stripField):
			return ""
		strip = self._stripIDAndStatesText
		if strip:
			strip += "\n"
		else:
			strip = ""
		return "%s%s" % (strip, self._cwaccess.stripField)

	def getCursorMBT(self):
		measure = ctypes.c_int()
		beat = ctypes.c_int()
		tick = ctypes.c_int()
		self._surface.GetTransportTime(SonarSurface.TIME_CURSOR, SonarSurface.TF_MBT, ctypes.pointer(measure), ctypes.pointer(beat), ctypes.pointer(tick))
		return measure.value, beat.value, tick.value

	def script_reportCursorTime(self, gesture):
		gesture.send()
		if scriptHandler.isScriptWaiting() or not self._surface:
			return
		measure, beat, tick = self.getCursorMBT()
		# Translators: Indicates a position in the project in measure/beat/tick format.
		# {measure} will be replaced with the measure number,
		# {beat} will be replaced with the beat number
		# and {tick} will be replaced with the tick number.
		ui.message(_("bar {measure}, beat {beat}, tick {tick}").format(
			measure=measure, beat=beat, tick=tick))

	def script_reportMarker(self, gesture):
		gesture.send()
		if scriptHandler.isScriptWaiting():
			return
		marker = self.appModule.getCurrentMarker()
		if not marker:
			# Translators: An error reported when the current marker cannot be retrieved.
			ui.message(_("Couldn't get current marker. Ensure the markers toolbar is enabled."))
			return
		ui.message(marker)

	def _toggleCurrentStripParam(self, mixerParam, onMessage, offMessage):
		if not self._surface:
			return
		stripNumber = self._cwaccess.stripIndex - 1
		param = self._surface.GetMixParam(self.surfaceStripType, stripNumber, mixerParam, 0)
		param = not param
		ui.message(onMessage if param else offMessage)
		self._surface.SetMixParam(self.surfaceStripType, stripNumber, mixerParam, 0, float(param), 0)

	def script_toggleScrubbing(self, gesture):
		if not self._surface:
			return
		state = not self._surface.GetTransportState(SonarSurface.TRANSPORT_STATE_SCRUB)
		self._surface.SetTransportState(SonarSurface.TRANSPORT_STATE_SCRUB, state)
		# Translators: Reported when scrubbing is started.
		ui.message(_("Scrubbing") if state else
			# Translators: Reported when scrubbing is stopped.
			_("Scrubbing stopped"))
	# Translators: Describes a command.
	script_toggleScrubbing.__doc__ = _("Toggles scrubbing.")

	def script_moveForward(self, gesture):
		if not self._surface:
			return
		self._surface.TimeAdvance(SonarSurface.TF_FRAMES, 3, 0, 0, False, False)
	# Translators: Describes a command.
	script_moveForward.__doc__ = _("Moves the cursor forward by the configured movement unit.")

	def script_moveBackward(self, gesture):
		if not self._surface:
			return
		self._surface.TimeAdvance(SonarSurface.TF_FRAMES, 3, 0, 0, True, True)
	# Translators: Describes a command.
	script_moveBackward.__doc__ = _("Moves the cursor backward by the configured movement unit.")

	def moveToField(self, text):
		if not waitForNoModifiers():
			return
		self._moveToFieldText = text
		self._moveToFieldNext(False)

	def _moveToFieldNext(self, stripIDChanged):
		if stripIDChanged:
			# Translators: An error reported when a strip field could not be found.
			ui.message(_("Field not found"))
			# Move back to the strip on which we started.
			KeyboardInputGesture.fromName("upArrow").send()
			self._moveToFieldText = None
			return
		if self._cwaccess.stripField.startswith(self._moveToFieldText):
			# Found it!
			self._moveToFieldText = None
			# Notify the user.
			self.event_cwaccessUpdate(False, True)
			return
		# Try the next field.
		# This method will be called again by the event when cwaccess gets updated.
		KeyboardInputGesture.fromName("rightArrow").send()

	__gestures = {
		"kb:control+pageDown": "reportCursorTime",
		"kb:control+pageUp": "reportCursorTime",
		"kb:control+home": "reportCursorTime",
		"kb:control+end": "reportCursorTime",
		"kb:control+shift+pageUp": "reportMarker",
		"kb:control+shift+pageDown": "reportMarker",
		"kb:control+b": "toggleScrubbing",
		"kb:control+j": "moveBackward",
		"kb:control+k": "moveForward",
	}

class TrackOrBusPane(CwaccessView):

	def script_moveToTrackInspector(self, gesture):
		obj = self.parent.parent.parent.parent.parent.previous.firstChild.firstChild
		if not obj:
			# Translators: Reported when the user attempts to move to the Track Inspector
			# but it is not enabled.
			ui.message(_("You must enable track inspector by pressing the letter i."))
			return
		obj.setFocus()
		# Set the window of origin for the TrackInspectorPane so that tab can return here.
		TrackInspectorPane._returnWindow = self.windowHandle

	def _getCurrentStripWindow(self):
		"""Obtain the window handle for the current strip.
		This must be implemented by subclasses.
		@return: The window handle for the current strip.
		@raise LookupError: If the window couldn't be found.
		"""
		raise NotImplementedError

	def script_activateStripContextMenu(self, gesture):
		try:
			window = self._getCurrentStripWindow()
		except NotImplementedError:
			return
		except LookupError:
			# Translators: Reported when the current strip couldn't be found on the screen,
			# which is necessary for some commands.
			ui.message(_("Couldn't find current strip on screen"))
			return
		winUser.PostMessage(window, WM_RBUTTONDOWN, 0, 0)
		winUser.PostMessage(window, WM_RBUTTONUP, 0, 0)

	def script_toggleCurrentStripMute(self, gesture):
		self._toggleCurrentStripParam(SonarSurface.MIX_PARAM_MUTE,
			# Translators: Reported when a strip is muted.
			_("muted"),
			# Translators: Reported when a strip is unmuted.
			_("unmuted"))

	def script_toggleCurrentStripSolo(self, gesture):
		self._toggleCurrentStripParam(SonarSurface.MIX_PARAM_SOLO,
			# Translators: Reported when a strip is soloed.
			_("soloed"),
			# Translators: Reported when a strip is unsoloed.
			_("unsoloed"))

	def script_changeTab(self, gesture):
		gesture.send()
		if scriptHandler.isScriptWaiting():
			return
		try:
			trackTabs = objFromWindow(findDescendantWindow(api.getFocusAncestors()[6].windowHandle, controlID=33049))
		except AttributeError:
			return
		# This control is only accessible with display model.
		# The active tab is bolded text.
		fields = trackTabs.makeTextInfo(textInfos.POSITION_ALL).getTextWithFields()
		boldFound = False
		for command in fields:
			if isinstance(command, textInfos.FieldCommand) and command.command == "formatChange" and command.field.get("bold"):
				boldFound = True
				continue
			elif isinstance(command, basestring) and boldFound:
				break
		else:
			return
		ui.message(speech.getSpeechTextForProperties(name=command, role=controlTypes.ROLE_TAB))

	def script_stripName(self, gesture):
		if scriptHandler.getLastScriptRepeatCount() == 1:
			# Jump to the name field.
			# The quickest way to do this is to click with the mouse.
			try:
				window = self._getCurrentStripWindow()
			except NotImplementedError:
				return
			except LookupError:
				ui.message(_("Couldn't find current strip on screen"))
				return
			l = winUser.MAKELONG(60, 10)
			winUser.PostMessage(window, WM_LBUTTONDOWN, 0, l)
			winUser.PostMessage(window, WM_LBUTTONUP, 0, l)
			return
		name = self.stripName
		if not name:
			return
		ui.message(name)
	# Translators: The description of a command.
	script_stripName.__doc__ = _("If pressed once, reports the name of the current strip. If pressed twice, moves to the strip name field.")

	def script_toggleEnvelopeOffsetMode(self, gesture):
		toggleToolbarControl(AUTOMATION_TOOLBAR, 5,
			# Translators: Reported for offset mode.
			_("Offset mode"),
			# Translators: Reported for envelope mode.
			_("Envelope mode"),
			gesture=gesture)

	def script_moveToFXBin(self, gesture):
		self.moveToField(_("FX bin"))
	# Translators: The description of a command.
	script_moveToFXBin.__doc__ = _("Moves to the FX bin on the current strip")

	def script_toggleAutomaticCrossfades(self, gesture):
		toggleToolbarControl(TRACK_VIEW_TOOLBAR, 17,
			# Translators: Reported when automatic crossfades are enabled.
			_("Automatic crossfades enabled"),
			# Translators: Reported when automatic crossfades are disabled.
			_("Automatic crossfades disabled"),
			gesture=gesture)

	__gestures = {
		"kb:tab": "moveToTrackInspector",
		"kb:control+applications": "activateStripContextMenu",
		"kb:control+.": "toggleCurrentStripMute",
		"kb:control+/": "toggleCurrentStripSolo",
		"kb:shift+leftArrow": "changeTab",
		"kb:shift+rightArrow": "changeTab",
		"kb:control+shift+n": "stripName",
		"kb:o": "toggleEnvelopeOffsetMode",
		"kb:control+shift+f": "moveToFXBin",
		"kb:x": "toggleAutomaticCrossfades",
	}

class TrackPane(TrackOrBusPane):
	# Translators: The label of the Track Pane.
	name = _("Track")
	surfaceStripType = SonarSurface.MIX_STRIP_TRACK

	def script_toggleCurrentStripArm(self, gesture):
		self._toggleCurrentStripParam(SonarSurface.MIX_PARAM_RECORD_ARM,
			# Translators: Reported when a track is armed.
			_("armed"),
			# Translators: Reported when a track is armed.
			_("unarmed"))

	def script_toggleCurrentStripInputEcho(self, gesture):
		self._toggleCurrentStripParam(SonarSurface.MIX_PARAM_INPUT_ECHO,
			# Translators: Reported when input echo for a track is turned on.
			_("input echo on"),
			# Translators: Reported when input echo for a track is turned on.
			_("input echo off"))

	def _getCurrentStripWindow(self):
		child = self.firstChild
		while child:
			if controlTypes.STATE_INVISIBLE in child.states:
				child = child.next
				continue
			x, y, w, h = child.location
			pixel = self.appModule._scrBmp.captureImage(x, y, 3, 3)[0][0]
			if pixel.rgbRed == pixel.rgbGreen:
				return child.windowHandle
			child = child.next
		raise LookupError("Couldn't find the current strip window")

	__gestures = {
		"kb:control+;": "toggleCurrentStripArm",
		"kb:control+'": "toggleCurrentStripInputEcho",
	}

class BusPane(TrackOrBusPane):
	# Translators: The label of the Bus Pane.
	name = _("Bus")
	surfaceStripType = SonarSurface.MIX_STRIP_BUS

	def _getCurrentStripWindow(self):
		# There doesn't seem to be any way to reliably get the window for the focused bus,
		# so just assume the windows are in sequential order.
		# Unfortunately, this isn't the case if buses have been reordered.
		return self.getChild(self._cwaccess.stripIndex - 1).windowHandle

class TrackInspectorPane(CwaccessView):
	# Translators: The label of the Track Inspector.
	name = _("Track Inspector")
	_returnWindow = None
	isStripIDRelevant = False
	forceUpdateKeys = ("upArrow", "downArrow")

	def script_returnToOrigin(self, gesture):
		if not self._returnWindow:
			return
		obj = objFromWindow(self._returnWindow)
		obj.setFocus()
		TrackInspectorPane._returnWindow = None

	__gestures = {
		"kb:tab": "returnToOrigin",
	}

class FXPresetToolbar(CwaccessView):
	# Translators: The label of the FX Preset toolbar.
	name = _("FX Preset")
	role = controlTypes.ROLE_TOOLBAR
	isStripIDRelevant = False

	def _get_value(self):
		# Scrap the strip ID.
		return "%s" % self._cwaccess.stripField if super(FXPresetToolbar, self).value else ""

class SynthRack(CwaccessView):
	# Translators: The label of the Synth Rack.
	name = _("Synth Rack")

class EventView(CwaccessView):
	# Translators: The label of the Event View.
	name = _("Event View")

class MatrixViewToolbar(CwaccessView):
	role = controlTypes.ROLE_TOOLBAR
	forceUpdateKeys = ()

class MatrixViewRowControls(CwaccessView):
	# Translators: The name of the Matrix View row controls.
	name = _("Row Controls")
	role = controlTypes.ROLE_GROUPING
	forceUpdateKeys = ()

class MatrixViewRow(CwaccessView):
	role = controlTypes.ROLE_TABLEROW
	forceUpdateKeys = ()

class HotSpotBase(NVDAObject):

	@classmethod
	def addHotSpot(cls, name, description, key, widthFrac, heightFrac):
		funcName = "script_%s" % name
		script = lambda self, gesture: cls._clickHotSpot(self, widthFrac, heightFrac)
		script.__doc__ = description
		script.__name__ = funcName
		script.canPropagate = True
		setattr(cls, funcName, script)
		gesturesAttr = "_%s__gestures" % cls.__name__
		try:
			gestures = getattr(cls, gesturesAttr)
		except AttributeError:
			gestures = {}
			setattr(cls, gesturesAttr, gestures)
		gestures["kb:%s" % key] = name

	def _clickHotSpot(self, widthFrac, heightFrac):
		l, t, w, h = self.location
		x = int(l + widthFrac * w)
		y = int(t + heightFrac * h)
		winUser.setCursorPos(x, y)
		winUser.mouse_event(winUser.MOUSEEVENTF_LEFTDOWN, 0, 0, None, None)
		winUser.mouse_event(winUser.MOUSEEVENTF_LEFTUP, 0, 0, None, None)

	def script_makeHotSpot(self, gesture):
		x, y = winUser.getCursorPos()
		l, t, w, h = self.location
		widthFrac = (x - l) / float(w)
		heightFrac = (y - t) / float(h)
		api.copyToClip(u'hs("name", _("desc"), "key", {w}, {h})'.format(w=widthFrac, h = heightFrac))
		# Translators: Reported when a hot spot definition has been copied to the clipboard
		# upon request from the user.
		# This command is only used by developers.
		ui.message(_("Hot spot definition copied to clipboard"))
	script_makeHotSpot.canPropagate = True

	__gestures = {
		"kb:NVDA+control+h": "makeHotSpot",
	}

class SfzPlugin(HotSpotBase):
	pass
hs = SfzPlugin.addHotSpot
# Translators: The name of the command which loads a sample in Sfz.
hs("loadSample", _("Load sample"), "alt+shift+l", 0.368932038835, 0.445454545455)

class DimensionProPlugin(HotSpotBase):
	pass
hs = DimensionProPlugin.addHotSpot
# Translators: The name of the command which opens the Program Browser in Dimension Pro.
hs("programBrowser", _("Program Browser"), "alt+shift+b", 0.0746460746461, 0.130208333333)

class Rapture32LEPlugin(HotSpotBase):
	pass
hs = Rapture32LEPlugin.addHotSpot
# Translators: The name of the command which opens the Program Browser in Raputre 32 LE.
hs("programBrowser", _("Program Browser"), "alt+shift+b", 0.112694300518, 0.166301969365)

class SessionDrummerPlugin(HotSpotBase):
	pass
hs = SessionDrummerPlugin.addHotSpot
# Translators: The name of the command which opens the Program Browser in Session Drummer.
hs("programBrowser", _("Program Browser"), "alt+shift+b", 0.130081300813, 0.83125)
del hs

PLUGIN_WINDOW_CLASSES = (
	# (windowTitleStart, class),
	("sfz", SfzPlugin),
	("Dimension Pro", DimensionProPlugin),
	("Rapture32LE", Rapture32LEPlugin),
	("SessionDrummer", SessionDrummerPlugin),
)

class ComplexToolbarControl(IAccessible.IAccessible):
	"""A control on a toolbar that requires more interaction than a single click.
	These controls have their own windows,
	but navigating next doesn't navigate to the simple controls (which are IAccessible child IDs).
	next is overridden here to fix this.
	"""

	def _get_next(self):
		obj = super(ComplexToolbarControl, self).next
		if not obj:
			return IAccessible.IAccessible(IAccessibleObject=self.parent.IAccessibleObject, IAccessibleChildID=1)
		return obj

class AppModule(appModuleHandler.AppModule):

	def __init__(self, *args, **kwargs):
		super(AppModule, self).__init__(*args, **kwargs)
		self._cwaccess = None
		self._surface = None
		self._scrBmp = screenBitmap.ScreenBitmap(1, 1)

		winVer = sys.getwindowsversion()
		if winVer.major == 5 and winVer.minor == 1:
			self._isWinXP = True
			# #1090586: In Windows XP, Sonar 8.5 crashes when closing the Markers View if an object is ever instantiated for its toolbar.
			# Unfortunately, a nameChange event is fired on that toolbar as soon as it opens, which guarantees the crash.
			# Stop this with a nasty monkey patch.
			self._origProcessGenericWinEvent = IAccessibleHandler.processGenericWinEvent
			IAccessibleHandler.processGenericWinEvent = self._processGenericWinEvent
		else:
			self._isWinXP = False

	def _processGenericWinEvent(self, eventID, window, objectID, childID):
		if eventID == winUser.EVENT_OBJECT_NAMECHANGE and winUser.getControlID(window) == 59392 and winUser.getClassName(window) == "ToolbarWindow32":
			return None
		return self._origProcessGenericWinEvent(eventID, window, objectID, childID)

	def terminate(self):
		if self._isWinXP:
			IAccessibleHandler.processGenericWinEvent = self._origProcessGenericWinEvent
			self._origProcessGenericWinEvent = None
		return super(AppModule, self).terminate()

	def chooseNVDAObjectOverlayClasses(self, obj, clsList):
		if obj.role == controlTypes.ROLE_WINDOW:
			if obj.windowClassName == "Afx:00400000:8:00000000:00000006:00000000":
				# Plugin window.
				name = obj.name
				if not name:
					return
				for title, objClass in PLUGIN_WINDOW_CLASSES:
					if name.startswith(title + " "):
						clsList.insert(0, objClass)
			parentWindow = winUser.getAncestor(obj.windowHandle, winUser.GA_PARENT)
			if parentWindow and winUser.getClassName(parentWindow) == "ToolbarWindow32":
				clsList.insert(0, ComplexToolbarControl)
			return

		if obj.windowClassName == "CWACCESS":
			clsList.insert(0, Cwaccess)
			return

		objClass = CwaccessView.getClass(obj)
		if objClass:
			if not self._cwaccess:
				self._initCwaccessAndSurface()
			try:
				clsList.remove(IAccessible.ContentGenericClient)
			except ValueError:
				pass
			clsList.insert(0, objClass)
			return

	def _initCwaccessAndSurface(self):
		handle = winUser.FindWindow(u"cwaccess", None)
		if handle:
			# Ensure that the event parameters are set correctly so that events will be fired on this instance.
			self._cwaccess = objFromWindow(handle)
		try:
			self._surface = comtypes.client.GetActiveObject("JSonar.ControlSurface", dynamic=True)
		except WindowsError:
			log.warning("JSonar control surface not installed")
			self._surface = None

	def getCurrentMarker(self):
		try:
			return objFromWindow(findDescendantWindow(api.getForegroundObject().windowHandle, controlID=1926, visible=True)).value
		except:
			log.debugWarning("", exc_info=True)
			return None

	def script_toggleArmAllTracks(self, gesture):
		if not waitForNoModifiers():
			return
		toggleToolbarControl(PLAYBACK_STATE_TOOLBAR, 3,
			# Translators: Reported when all tracks are armed.
			_("Armed all tracks"),
			# Translators: Reported when all tracks are unarmed.
			_("Unarmed all tracks"))
	# Translators: The description of a command.
	script_toggleArmAllTracks.__doc__ = _("Toggles arm for all tracks")

	def script_test(self, gesture):
		pass

	__gestures = {
		"kb:NVDA+control+t": "test",
		"kb:alt+shift+;": "toggleArmAllTracks",
	}
